/*------------------------------------------------------------------------------
 * MDK Middleware - Component ::Network
 * Copyright (c) 2016 Real Time Logic. All rights reserved.
 *------------------------------------------------------------------------------
 * Name:    SMQ.c
 * Purpose: LED Client example using BSD sockets
 *----------------------------------------------------------------------------*/

#include "cmsis_os.h"                   /* CMSIS RTOS definitions             */
#include "rl_net.h"                     /* Network definitions                */

#include "Board_LED.h"

#include "board.h"                      /* BSP definitions                    */

#include <ledctrl.h>
#include <stdio.h>


static const LedInfo ledInfo[] = {
   {
      "LED 1",
      LedColor_red,
      1
   },
   {
      "LED 2",
      LedColor_green,
      2
   },
   {
      "LED 3",
      LedColor_blue,
      3
   }
};

#define LEDSLEN (sizeof(ledInfo) / sizeof(ledInfo[0]))

static int leds[LEDSLEN];

/* Returns the LED on/off state for led with ID 'ledId'.
 */
int getLedState(int ledId)
{
   if(ledId >= 1 && ledId <= LEDSLEN)
      return leds[ledId-1];
   return 0;
}

/*
  Return an array of LedInfo (struct). Each element in the array
  provides information for one LED. The 'len' argument must be set by
  function getLedInfo.  The out argument 'en' specifies the length of
  the returned array, that is, number of LEDs in the device.  Each LED
  has a name, color, and ID. The ID, which provides information about
  which LED to turn on/off, is used by control messages sent between
  device code and UI clients. The IDs for a four LED device can for
  example be 1,2,3,4.
*/
const LedInfo* getLedInfo(int* len)
{
   *len = LEDSLEN;
   return ledInfo;
}


/* Returns the name of this device. The name is presented by UI
   clients such as browsers.
*/
const char* getDevName(void)
{
   static char name[30]={0};
   if(!*name)
      snprintf(name,sizeof(name),"ARM MDK SharkSSL-Lite: %s",netSYS_GetHostName());
   return name;
}


/* Blink all LEDs
 */
static void blinkAll(int fast)
{
   int i;
   for(i = 0 ; i < LEDSLEN ; i++) setLed(ledInfo[i].id, FALSE);
   for(i = 0 ; i < LEDSLEN ; i++)
   {
      setLed(ledInfo[i].id, TRUE);
      osDelay(fast ? 120: 400);
   }
   for(i = 0 ; i < LEDSLEN ; i++)
   {
      setLed(ledInfo[i].id, FALSE);
      osDelay(fast ? 120: 400);
   }
}

/* Blink LED with ID 'led'
 */
static void blink(int led)
{
   int i;
   for(i = 0 ; i < 30; i++)
   {
      setLed(led, 1);
      osDelay(120);
      setLed(led, 0);
      osDelay(120);
   }
}

/* Command sent by UI client to turn LED with ID on or off. This
   function must set the LED to on if 'on' is TRUE and off if 'on' is FALSE.
*/
int setLed(int ledId, int on)
{
   if(ledId >= 1 && ledId <= LEDSLEN)
   {
      leds[ledId-1] = on;
      if(on) 
         LED_On(ledId-1);
      else
         LED_Off(ledId-1);
      return 0;
   }
   return -1;
}


/*
  An optional function that enables LEDs to be set directly by the
  device. This function is typically used by devices that include one
  or more buttons. A button click may for example turn on a specific
  LED. The function is called at intervals (polled) by the LED device
  code. The function may for example detect a button click and return
  the information to the caller. Arguments 'ledId' and 'on' are out
  arguments, where 'ledId' is set to the LED ID and 'on' is set to
  TRUE for on and FALSE for off. The function must return TRUE (a non
  zero value) if the LED is to be set on/off and zero on no
  change. Create an empty function returning zero if you do not plan
  on implementing this feature.
*/
int setLedFromDevice(int* ledId, int* on)
{
   return FALSE;
}

void setProgramStatus(ProgramStatus s)
{
   int i;
   for(i = 0 ; i < LEDSLEN ; i++) setLed(ledInfo[i].id, FALSE);
   switch(s)
   {
      case ProgramStatus_Restarting:
         blinkAll(FALSE);
      case ProgramStatus_Starting:
         blinkAll(TRUE);
         break;

      case ProgramStatus_Connecting:
         setLed(ledInfo[1].id, TRUE);
         break;

      case ProgramStatus_SslHandshake:
         setLed(ledInfo[2].id, TRUE);
         break;

      case ProgramStatus_DeviceReady:
         for(i = 0 ; i < LEDSLEN ; i++) setLed(ledInfo[i].id, TRUE);
         osDelay(150);
         for(i = 0 ; i < LEDSLEN ; i++) setLed(ledInfo[i].id, FALSE);
         break;

      case ProgramStatus_CloseCommandReceived:
         blink(ledInfo[2].id);
         break;

      default:
         blinkAll(FALSE);
         switch(s)
         {
            case ProgramStatus_SocketError:
            case ProgramStatus_DnsError:
            case ProgramStatus_WebServiceNotAvailError:
            case ProgramStatus_InvalidCommandError:
            case ProgramStatus_PongResponseError:
               blink(ledInfo[1].id);
               break;

            case ProgramStatus_ConnectionError:
               blink(ledInfo[2].id);
               break;

            case ProgramStatus_CertificateNotTrustedError:
            case ProgramStatus_SslHandshakeError:
               blink(ledInfo[2].id);
               break;

            case ProgramStatus_MemoryError:
               for(i = 0 ; i < LEDSLEN ; i++) blink(ledInfo[i].id);
               break;
         }
   }
}


/* Required by SMQ examples.
   The unique ID is used when calling the SMQ constructor. The
   unique ID is typically set to the MAC address. See the SMQ
   documentation for details:
   https://realtimelogic.com/ba/doc/en/C/shark/structSharkMQ.html
*/
int getUniqueId(const char** id)
{
   static uint8_t addr[NET_ADDR_ETH_LEN];
   netIF_GetOption(
      NET_IF_CLASS_ETH, netIF_OptionMAC_Address, addr, NET_ADDR_ETH_LEN);
   *id = (char*)addr;
   return NET_ADDR_ETH_LEN;
}


static void wait4DHCP(void)
{
   uint8_t ip_addr[NET_ADDR_IP6_LEN];
   ip_addr[0]=0;
   do
   {
      osDelay(300);
      netIF_GetOption(NET_IF_CLASS_ETH, netIF_OptionIP4_Address,
                      ip_addr, sizeof(ip_addr));
   } while(!*ip_addr);
}

/*----------------------------------------------------------------------------
  Main Thread 'main': Run Network
  *---------------------------------------------------------------------------*/
int main (void) {
   hardware_init();
   LED_Initialize();
   netInitialize();
   wait4DHCP();
   umm_init(); /* Init UMM memory allocator used by SharkSSL. */
   mainTask(0); // SMQ LED example: m2m-led.c
}
