/*------------------------------------------------------------------------------
 * MDK Middleware - Component ::Network
 * Copyright (c) 2016 Real Time Logic. All rights reserved.
 *------------------------------------------------------------------------------
 * Name:    SMQ.c
 * Purpose: LED Client example using BSD sockets
 *----------------------------------------------------------------------------*/

#include <string.h>
#include "cmsis_os.h"                   // ARM::CMSIS:RTOS:Keil RTX
#include "rl_net.h"                     // Keil.MDK-Pro::Network:CORE

#include "stm32f7xx_hal.h"              // Keil::Device:STM32Cube HAL:Common

#include "Board_LED.h"                  // ::Board Support:LED
#include "Board_GLCD.h"                 // ::Board Support:Graphic LCD
#include "GLCD_Config.h"                // Keil.STM32F746G-Discovery::Board Support:Graphic LCD

#include <ledctrl.h>
#include <stdio.h>


extern GLCD_FONT GLCD_Font_16x24;


#ifdef RTE_CMSIS_RTOS_RTX
extern uint32_t os_time;

uint32_t HAL_GetTick(void) {
   return os_time; 
}
#endif

/**
 * System Clock Configuration
 *   System Clock source            = PLL (HSE)
 *   SYSCLK(Hz)                     = 216000000
 *   HCLK(Hz)                       = 216000000
 *   AHB Prescaler                  = 1
 *   APB1 Prescaler                 = 4
 *   APB2 Prescaler                 = 2
 *   HSE Frequency(Hz)              = 25000000
 *   PLL_M                          = 25
 *   PLL_N                          = 432
 *   PLL_P                          = 2
 *   PLL_Q                          = 9
 *   VDD(V)                         = 3.3
 *   Main regulator output voltage  = Scale1 mode
 *   Flash Latency(WS)              = 7
 */
static void SystemClock_Config (void) {
   RCC_ClkInitTypeDef RCC_ClkInitStruct;
   RCC_OscInitTypeDef RCC_OscInitStruct;

   /* Enable HSE Oscillator and activate PLL with HSE as source */
   RCC_OscInitStruct.OscillatorType = RCC_OSCILLATORTYPE_HSE;
   RCC_OscInitStruct.HSEState = RCC_HSE_ON;
   RCC_OscInitStruct.HSIState = RCC_HSI_OFF;
   RCC_OscInitStruct.PLL.PLLState = RCC_PLL_ON;
   RCC_OscInitStruct.PLL.PLLSource = RCC_PLLSOURCE_HSE;
   RCC_OscInitStruct.PLL.PLLM = 25;
   RCC_OscInitStruct.PLL.PLLN = 432;  
   RCC_OscInitStruct.PLL.PLLP = RCC_PLLP_DIV2;
   RCC_OscInitStruct.PLL.PLLQ = 9;
   HAL_RCC_OscConfig(&RCC_OscInitStruct);

   /* Activate the OverDrive to reach the 216 MHz Frequency */
   HAL_PWREx_EnableOverDrive();
  
   /* Select PLL as system clock source and configure the HCLK, PCLK1 and PCLK2 clocks dividers */
   RCC_ClkInitStruct.ClockType = (RCC_CLOCKTYPE_SYSCLK | RCC_CLOCKTYPE_HCLK | RCC_CLOCKTYPE_PCLK1 | RCC_CLOCKTYPE_PCLK2);
   RCC_ClkInitStruct.SYSCLKSource = RCC_SYSCLKSOURCE_PLLCLK;
   RCC_ClkInitStruct.AHBCLKDivider = RCC_SYSCLK_DIV1;
   RCC_ClkInitStruct.APB1CLKDivider = RCC_HCLK_DIV4;  
   RCC_ClkInitStruct.APB2CLKDivider = RCC_HCLK_DIV2;  
   HAL_RCC_ClockConfig(&RCC_ClkInitStruct, FLASH_LATENCY_7);
}

/**
 * Configure the MPU attributes as Write Through for SRAM1/2
 *   The Base Address is 0x20010000 since this memory interface is the AXI.
 *   The Region Size is 256KB, it is related to SRAM1 and SRAM2 memory size.
 */
static void MPU_Config (void) {
   MPU_Region_InitTypeDef MPU_InitStruct;
  
   /* Disable the MPU */
   HAL_MPU_Disable();

   /* Configure the MPU attributes as WT for SRAM */
   MPU_InitStruct.Enable = MPU_REGION_ENABLE;
   MPU_InitStruct.BaseAddress = 0x20010000;
   MPU_InitStruct.Size = MPU_REGION_SIZE_256KB;
   MPU_InitStruct.AccessPermission = MPU_REGION_FULL_ACCESS;
   MPU_InitStruct.IsBufferable = MPU_ACCESS_NOT_BUFFERABLE;
   MPU_InitStruct.IsCacheable = MPU_ACCESS_CACHEABLE;
   MPU_InitStruct.IsShareable = MPU_ACCESS_NOT_SHAREABLE;
   MPU_InitStruct.Number = MPU_REGION_NUMBER0;
   MPU_InitStruct.TypeExtField = MPU_TEX_LEVEL0;
   MPU_InitStruct.SubRegionDisable = 0x00;
   MPU_InitStruct.DisableExec = MPU_INSTRUCTION_ACCESS_ENABLE;

   HAL_MPU_ConfigRegion(&MPU_InitStruct);

   /* Enable the MPU */
   HAL_MPU_Enable(MPU_PRIVILEGED_DEFAULT);
}

/**
 * CPU L1-Cache enable
 */
static void CPU_CACHE_Enable (void) {

   /* Enable I-Cache */
   SCB_EnableICache();

   /* Enable D-Cache */
   SCB_EnableDCache();
}

/* Initialize display */
static void init_display () {
   GLCD_Initialize         ();
   GLCD_SetBackgroundColor (GLCD_COLOR_BLUE);
   GLCD_SetForegroundColor (GLCD_COLOR_WHITE);
   GLCD_ClearScreen        ();
   GLCD_SetFont            (&GLCD_Font_16x24);
   GLCD_DrawString         (5*16, 1*24, "       SMQ       ");
   GLCD_DrawString         (5*16, 2*24, "Secure LED Example");
}

/// Print IP addresses
static void print_address (void) {
   static uint8_t ip_addr[NET_ADDR_IP6_LEN];
   static char    ip_ascii[40];
   static char    buf[24];

   /* Print IPv4 network address */
   netIF_GetOption (NET_IF_CLASS_ETH, netIF_OptionIP4_Address, ip_addr, sizeof(ip_addr));
   netIP_ntoa (NET_ADDR_IP4, ip_addr, ip_ascii, sizeof(ip_ascii));

   sprintf (buf, "IP4:%-15s", ip_ascii);
   GLCD_DrawString (5*16, 4*24, buf);

#if 0
   /* Print IPv6 network address */
   netIF_GetOption (NET_IF_CLASS_ETH, netIF_OptionIP6_LinkLocalAddress, ip_addr, sizeof(ip_addr));
   netIP_ntoa(NET_ADDR_IP6, ip_addr, ip_ascii, sizeof(ip_ascii));
   sprintf (buf, "IP6:%.16s", ip_ascii);
   GLCD_DrawString ( 5*16, 5*24, buf);
   sprintf (buf, "%s", ip_ascii+16);
   GLCD_DrawString (15*16, 6*24, buf);
#endif

}

static void printStatus(const char* str)
{
   GLCD_DrawString (5*16, 7*24, str);
}



static const LedInfo ledInfo[] = {
   {
      "LED 1",
      LedColor_green,
      1
   }
};

#define LEDSLEN (sizeof(ledInfo) / sizeof(ledInfo[0]))

static int leds[LEDSLEN];

/* Returns the LED on/off state for led with ID 'ledId'.
 */
int getLedState(int ledId)
{
   if(ledId >= 1 && ledId <= LEDSLEN)
      return leds[ledId-1];
   return 0;
}

/*
  Return an array of LedInfo (struct). Each element in the array
  provides information for one LED. The 'len' argument must be set by
  function getLedInfo.  The out argument 'en' specifies the length of
  the returned array, that is, number of LEDs in the device.  Each LED
  has a name, color, and ID. The ID, which provides information about
  which LED to turn on/off, is used by control messages sent between
  device code and UI clients. The IDs for a four LED device can for
  example be 1,2,3,4.
*/
const LedInfo* getLedInfo(int* len)
{
   *len = LEDSLEN;
   return ledInfo;
}


/* Returns the name of this device. The name is presented by UI
   clients such as browsers.
*/
const char* getDevName(void)
{
   static char name[30]={0};
   if(!*name)
      snprintf(name,sizeof(name),"ARM MDK SharkSSL-Lite: %s",netSYS_GetHostName());
   return name;
}


/* Command sent by UI client to turn LED with ID on or off. This
   function must set the LED to on if 'on' is TRUE and off if 'on' is FALSE.
*/
int setLed(int ledId, int on)
{
   if(ledId >= 1 && ledId <= LEDSLEN)
   {
      leds[ledId-1] = on;
      if(on) 
         LED_On(ledId-1);
      else
         LED_Off(ledId-1);
      return 0;
   }
   return -1;
}


/*
  An optional function that enables LEDs to be set directly by the
  device. This function is typically used by devices that include one
  or more buttons. A button click may for example turn on a specific
  LED. The function is called at intervals (polled) by the LED device
  code. The function may for example detect a button click and return
  the information to the caller. Arguments 'ledId' and 'on' are out
  arguments, where 'ledId' is set to the LED ID and 'on' is set to
  TRUE for on and FALSE for off. The function must return TRUE (a non
  zero value) if the LED is to be set on/off and zero on no
  change. Create an empty function returning zero if you do not plan
  on implementing this feature.
*/
int setLedFromDevice(int* ledId, int* on)
{
   return FALSE;
}

void setProgramStatus(ProgramStatus s)
{
   const char* msg;
   switch(s)
   {
      case ProgramStatus_Starting: msg="Starting"; break;
      case ProgramStatus_Restarting: msg="Restarting"; break;
      case ProgramStatus_Connecting: msg="Connecting"; break;
      case ProgramStatus_SslHandshake: msg="SslHandshake"; break;
      case ProgramStatus_DeviceReady: msg="DeviceReady"; break;
      case ProgramStatus_CloseCommandReceived: msg="CloseCommandReceived"; break;
      case ProgramStatus_SocketError: msg="SocketError"; break;
      case ProgramStatus_DnsError: msg="DnsError"; break;
      case ProgramStatus_ConnectionError: msg="ConnectionError"; break;
      case ProgramStatus_CertificateNotTrustedError: msg="CertificateNotTrustedError"; break;
      case ProgramStatus_SslHandshakeError: msg="SslHandshakeError"; break;
      case ProgramStatus_WebServiceNotAvailError: msg="WebServiceNotAvailError"; break;
      case ProgramStatus_PongResponseError: msg="PongResponseError"; break;
      case ProgramStatus_InvalidCommandError: msg="InvalidCommandError"; break;
      default:
         msg="MemoryError"; break;
   }
   printStatus(msg);
   switch(s)
   {
      case ProgramStatus_CloseCommandReceived:
      case ProgramStatus_SocketError:
      case ProgramStatus_DnsError:
      case ProgramStatus_ConnectionError:
      case ProgramStatus_CertificateNotTrustedError:
      case ProgramStatus_SslHandshakeError:
      case ProgramStatus_WebServiceNotAvailError:
      case ProgramStatus_PongResponseError:
      case ProgramStatus_InvalidCommandError:
         osDelay(5000); /* so user gets time to read error msg */
   }
}



/* Required by SMQ examples.
   The unique ID is used when calling the SMQ constructor. The
   unique ID is typically set to the MAC address. See the SMQ
   documentation for details:
   https://realtimelogic.com/ba/doc/en/C/shark/structSharkMQ.html
*/
int getUniqueId(const char** id)
{
   static uint8_t addr[NET_ADDR_ETH_LEN];
   netIF_GetOption(
      NET_IF_CLASS_ETH, netIF_OptionMAC_Address, addr, NET_ADDR_ETH_LEN);
   *id = (char*)addr;
   return NET_ADDR_ETH_LEN;
}


static void wait4DHCP(void)
{
   uint8_t ip_addr[NET_ADDR_IP6_LEN];
   ip_addr[0]=0;
   do
   {
      osDelay(300);
      netIF_GetOption(NET_IF_CLASS_ETH, netIF_OptionIP4_Address,
                      ip_addr, sizeof(ip_addr));
   } while(!*ip_addr);
}


/*----------------------------------------------------------------------------
  Main Thread 'main': Run Network
  *---------------------------------------------------------------------------*/
int main (void) {

   MPU_Config();                             /* Configure the MPU              */
   CPU_CACHE_Enable();                       /* Enable the CPU Cache           */

   HAL_Init();                               /* Initialize the HAL Library     */
   SystemClock_Config();                     /* Configure the System Clock     */

   LED_Initialize ();
   init_display   ();
  
   netInitialize ();

   printStatus("DHCP.....");
   wait4DHCP();
   print_address();

   umm_init(); /* Init UMM memory allocator used by SharkSSL. */
   mainTask(0); // SMQ LED example: m2m-led.c
}
